/*
 * Program        : fpt_mpis.c
 *
 * Objectives     : to implement the functions used in FP-MPIS-tree
 *
 * Assumptions    : Item ID in dataset file should start from zero.
 *                  I.e. if there are NUM items in the dataset, 
 *                  the item ID should be from 0 to NUM-1.
 *
 * Variations     : 
 *
 * Platform       : SunOS/Linux
 * Compiler       : gcc
 *
 * Programmer     : Raymond Chi-Wing Wong
 * Creation Date  : Oct, 2002
 * Last Update    : Jun 9, 2004
 *
 * Reference      :
 *   1. J. Han, J. Pei, and Y. Yin, 
 *      "Mining Frequent Patterns without Candidate Generation", 
 *      Proc. ACM SIGMOD 2000.
 *   2. R. C.-W. Wong, A. W.-C. Fu and K. Wang
 *      "MPIS: Maximal-Profit Item Selection with Cross-Selling Considerations"
 *      The 2003 IEEE International Conference on Data Mining (ICDM)
 *      
 */

#include "fpt_mpis.h"



/*
 * Function  : isContainCase
 * Purpose   : to return which case the treansaction falls
 * Parameters: 
 *   1. transaction: transaction
 *   2. isChoose: an array of items indicating whether it is selected or not
 * NOTES:
 *   1. return:
 *       2 : all selected
 *       1 : some selected and some unselected
 *       0 : all unselected  
 */

int isContainCase(transPtr transaction, int *isChoose)
{
	int returnValue = 0;

	int i;

	int isAll;

	for (i = 0; i < transaction->noOfItem; i++)
	{
		if (isChoose[transaction->item[i]] == 1)
		{
			returnValue = 1;
			
			if (i == 0)
			{
				break;
			}
			else
			{
				return returnValue;
			}
		}
	}

	// check for all selected items
	if (returnValue == 1)
	{
		isAll = 1;
		for (i = 0; i < transaction->noOfItem; i++)
		{
			if (isChoose[transaction->item[i]] == 0)
			{
				isAll = 0;
				return 1;
			}
		}

		if (isAll == 1)
		{
			return 2;
		}
	}

	return returnValue;
}


/*
 * Function  : isContainAddRemove
 * Purpose   : to return whether the transaction contains itemAdd or itemRemove
 * Parameters: 
 *   1. transaction: a transaction
 *   2. itemAdd: item just added
 *   3. itemRemove: item just removed
 * NOTES:
 *   1. return:
 *       1 : exist
 *       0 : not exist
 */

int isContainAddRemove(transPtr transaction, int itemAdd, int itemRemove)
{
	int returnValue = 0;

	int i;
	int itemNo;

	for (i = 0; i < transaction->noOfItem; i++)
	{
		itemNo = transaction->item[i];

		if ((itemNo == itemAdd) || (itemNo == itemRemove))
		{
			return 1;
		}
	}

	return returnValue;
}


/*
 * Function  : find_no (dummy)
 * Purpose   : to find whether the children is null or not
 * Parameters: 
 *   1. node: the node to be found whether the children is null or not
 */
void find_no(FPTreeNode node)
{
	childLink nextChild;

	if (node == NULL)
	{
		printf(" node is NULL!\n");
		exit(1);
	}

	printf("Item %d - Count %d \n", node->item, node->count);

	nextChild = node->children;
	
	while (nextChild != NULL)
	{
		find_no(nextChild->node);

		nextChild = nextChild->next;
	}
	printf("-----------End of item %d--\n", node->item);
}


/*
 * Function  : fp_copyCount
 * Purpose   : to copy the count from "count" to "numPath"
 * Parameters: 
 *   1. node: the node to be copied with the "count"
 */
void fp_copyCount(FPTreeNode node)
{
	childLink nextChild;

	if (node == NULL)
	{
		return;
	}

	// copy
	node->numPath = node->count;

	// explore children
	nextChild = node->children;
	while (nextChild != NULL)
	{
		fp_copyCount(nextChild->node);		
			
		nextChild = nextChild->next;
	}	
}


/*
 * Function  : fp_doTransaction
 * Purpose   : to calculate the profit of the transaction
 * Parameters: 
 *   1. returnValue: the profit of the transaction
 *   2. transaction: a transaction
 *   3. freqItemP: the ordered freq. items
 *   4. indexList: the list of index of freq. item
 *   5. node: the root node of the FP-tree
 *   6. headerTabelLink: the header table of FP-tree
 *   7. fpMap: the mapping table from the header table to item no.
 *   8. isChoose: the array indicating whether the item is chosen or not
 *   9. profit: the profit of items
 *  10. noOfContain: no. of selected items in the transaction
 */

void fp_doTransaction(double *returnValue, transPtr transaction, int *freqItemP, int *indexList, FPTreeNode node, FPTreeNode *headerTableLink, short *fpMap, int *isChoose, int *profit, int noOfContain)
{
	int flag;
	
	int j;
	double temp;
	
	int path;
	int itemNo;

	int count;

	flag = isContainCase(transaction, isChoose);
	
	if (flag == 0)
	{
		// all unselected items
		
		// do nothing
	}
	else if (flag == 2)
	{
		// all selected items
		
		temp = 0;
		for (j = 0; j < transaction->noOfItem; j++)
		{
			itemNo = transaction->item[j];
			
			temp += profit[itemNo];
		}
		temp *= noOfContain;
		
		(*returnValue) += temp;
	}
	else
	{
		// flag == 1
		// some selected and some unselected
		
		count = 0;
		path = 0;
		
		for (j = 0; j < transaction->noOfItem; j++)
		{
			itemNo = transaction->item[j];
			
			freqItemP[count] = itemNo;
			indexList[count] = (int)fpMap[itemNo];
			count++;
		}
		
		q_sortA(indexList, freqItemP, 0, count-1, count);
		
		insert_tree(freqItemP, indexList, noOfContain, 0, count, node, headerTableLink, &path);
	}
}


/*
 * Function  : doCreateFP
 * Purpose   : to create a new FP-tree
 * Parameters: 
 *   1. returnValue: the profit to be calculated
 *   2. freqItemP: the ordered freq. items
 *   3. indexList: the list of index of freq. item
 *   4. node: the root node of the FP-tree
 *   5. fpMap: the mapping table from the header table to item no.
 *   6. isChoose: the array indicating whether the item is chosen or not
 *   7. profit: the profit of items
 */
// for normal
void doCreateFP(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, short *fpMap, int *isChoose, int *profit)
{
	transPtr transaction;

	int noOfContain;
	int count;

	FPTreeNode parent;



	// check whether the node's count is zero
	if (node->numPath == 0)
	{
		return;
	}
	else
	{
		// create trans
		transaction = (transPtr) malloc(sizeof(trans));
		transaction->item = (int *) malloc(sizeof(int)*noOfItem);
		
		noOfContain = node->numPath;
		
		count = 0;
		node->numPath -= noOfContain;
		transaction->item[count] = node->item;
		count++;
		
		parent = node->parent;
		
		
		while ((parent != NULL) && (parent->parent != NULL))
		{
			parent->numPath -= noOfContain;
			transaction->item[count] = parent->item;
			count++;
			
			parent = parent->parent;
		}
		
		transaction->noOfItem = count;
		
		// just create a transaction
		

		fp_doTransaction(returnValue, transaction, freqItemP, indexList, rootSet, headerTableLinkSet, fpMap, isChoose, profit, noOfContain);
		
		
		
		free(transaction->item);
		free(transaction);
	}
}


/*
 * Function  : doCreateFP2
 * Purpose   : to create a new FP-tree (for the purpose of the calculation of H)
 * Parameters: 
 *   1. returnValue: the profit to be calculated
 *   2. freqItemP: the ordered freq. items
 *   3. indexList: the list of index of freq. item
 *   4. node: the root node of the FP-tree
 *   5. fpMap: the mapping table from the header table to item no.
 *   6. isChoose: the array indicating whether the item is chosen or not
 *   7. profit: the profit of items
 *   8. revNo: the reverse mapping of fpMap
 *   9. itemSpecial: the item to be calculated in H
 */
// for H
void doCreateFP2(double *returnValue, double *returnValue2, int *freqItemP, int *indexList, FPTreeNode node, short *fpMap, int *isChoose, int *profit, double *revNo, int itemSpecial)
{
	transPtr transaction;

	int noOfContain;
	int count;

	FPTreeNode parent;

	int flag;

	int j;
	double temp;

	int path;
	int itemNo;

	int isExist;

	// check whether the node's count is zero
	if (node->numPath == 0)
	{
		return;
	}
	else
	{
		// create trans
		transaction = (transPtr) malloc(sizeof(trans));
		transaction->item = (int *) malloc(sizeof(int)*noOfItem);
		
		noOfContain = node->numPath;
		
		count = 0;
		node->numPath -= noOfContain;
		transaction->item[count] = node->item;
		count++;
		
		parent = node->parent;
		
		
		while ((parent != NULL) && (parent->parent != NULL))
		{
			parent->numPath -= noOfContain;
			transaction->item[count] = parent->item;
			count++;
			
			parent = parent->parent;
		}
		
		transaction->noOfItem = count;
		
		// just create a transaction
		
		flag = isContainCase(transaction, isChoose);
		
		if (flag == 0)
		{
			// all unselected items
			
			// do nothing
		}
		else
		{
			// check whether the trans contain specialItem
			isExist = 0;
			for (j = 0; j < transaction->noOfItem; j++)
			{
				if (transaction->item[j] == itemSpecial)
				{
					isExist = 1;
					break;
				}
			}

			if (flag == 2)
			{
				// all selected items
				
				temp = 0;
				for (j = 0; j < transaction->noOfItem; j++)
				{
					itemNo = transaction->item[j];
					
					temp += profit[itemNo];
				}
				temp *= noOfContain;
				
				(*returnValue) += temp;

				if (isExist == 0)
				{
					(*returnValue2) += temp;
				}
				else
				{
					(*returnValue2) += temp;

					for (j = 0; j < transaction->noOfItem; j++)
					{
						itemNo = transaction->item[j];
						if (itemNo == itemSpecial)
						{
							(*returnValue2) -= profit[itemSpecial];
						}
						else
						{
							(*returnValue2) -= profit[itemNo]*getMatrix(itemSpecial, itemNo)*revNo[itemNo];
						}
					}
				}
			}
			else
			{
				// flag == 1
				// some selected and some unselected
				
				count = 0;
				path = 0;
				
				for (j = 0; j < transaction->noOfItem; j++)
				{
					itemNo = transaction->item[j];
					
					freqItemP[count] = itemNo;
					indexList[count] = (int)fpMap[itemNo];
					count++;
				}
				
				q_sortA(indexList, freqItemP, 0, count-1, count);
				
				insert_tree(freqItemP, indexList, noOfContain, 0, count, rootSet, headerTableLinkSet, &path);
			}
		}
		
		free(transaction->item);
		free(transaction);
	}
}


/*
 * Function  : doCreateFP_diff
 * Purpose   : to calculate the FP-tree with difference profit
 * Parameters: 
 *   1. returnValue: profit 
 *   2. returnValue2: profit difference
 *   3. freqItemP: the ordered freq. items
 *   4. indexList: the list of index of freq. item
 *   5. node: the root node the FP-tree
 *   6. fpMap: the mapping table from the header table to item no.
 *   7. fpMap2: the mapping table from the header table to item no.
 *   8. isChoose: the array indicating whether the item is chosen or not
 *   9. profit: the profit of items
 *  10. itemAdd: the item just added
 *  11. itemRemove: the item just removed
 */
// for update
void doCreateFP_diff(double *returnValue, double *returnValue2, int *freqItemP, int *indexList, FPTreeNode node, short *fpMap, short *fpMap2, int *isChoose, int *profit, int itemAdd, int itemRemove)
{
	transPtr transaction;

	int noOfContain;

	int isContain;

	int *isChoose2;

	int count;

	FPTreeNode parent;



	// check whether the node's count is zero
	if (node->numPath == 0)
	{
		return;
	}
	else
	{
		isChoose2 = (int *) malloc(sizeof(int)*noOfItem);
		memcpy(isChoose2, isChoose, sizeof(int)*noOfItem);

		// create trans
		transaction = (transPtr) malloc(sizeof(trans));
		transaction->item = (int *) malloc(sizeof(int)*noOfItem);
		
		noOfContain = node->numPath;
		
		count = 0;
		node->numPath -= noOfContain;
		transaction->item[count] = node->item;
		count++;
		
		parent = node->parent;
		
		
		while ((parent != NULL) && (parent->parent != NULL))
		{
			parent->numPath -= noOfContain;
			transaction->item[count] = parent->item;
			count++;
			
			parent = parent->parent;
		}
		
		transaction->noOfItem = count;
		
		// just create a transaction

		// check whether contains itemAdd or itemRemove
		isContain = isContainAddRemove(transaction, itemAdd, itemRemove);
		if (isContain == 1)
		{
			
			// ***** For S U itemAdd
			isChoose2[itemAdd] = 1;
			isChoose2[itemRemove] = 0;

			// do transaction
			fp_doTransaction(returnValue, transaction, freqItemP, indexList, rootSet, headerTableLinkSet, fpMap, isChoose2, profit, noOfContain);


			// ***** For S U itemRemove
			isChoose2[itemAdd] = 0;
			isChoose2[itemRemove] = 1;

			// do transaction
			fp_doTransaction(returnValue2, transaction, freqItemP, indexList, rootSet2, headerTableLinkSet2, fpMap2, isChoose2, profit, noOfContain);
		}
		
		free(transaction->item);
		free(transaction);
		free(isChoose2);
	}
}


/*
 * Function  : fp_scanTrans
 * Purpose   : to scan the transactions in the FP-tree
 * Parameters: 
 *   1. returnValue: the profit of the transaction
 *   2. freqItemP: the ordered freq. items
 *   3. indexList: the list of index of freq. item
 *   4. node: the root node of the FP-tree
 *   5. data: the data in the database
 *   6. fpMap: the mapping table from the header table to item no.
 *   7. isChoose: the array indicating whether the item is chosen or not
 *   8. profit: the profit of items
 */
// for normal
void fp_scanTrans(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, int *isChoose, int *profit)
{
	
	//int itemNo;

	childLink nextChild;
	
	if (node == NULL)
	{
		return;
	}


	if (node->children == NULL)
	{
		// this node has no children
		// (i.e. this node is leaf)
		
		
		//doCreateFP(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, short *fpMap, int *isChoose, int *profit)
		doCreateFP(returnValue, freqItemP, indexList, node, fpMap, isChoose, profit);
	}
	else
	{
		// this node has children
		// (need to explore the children first
		//  by using depth-first search)

		// explore children
		nextChild = node->children;
		while (nextChild != NULL)
		{
			//fp_scanTrans(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, int *isChoose, int *profit)
			fp_scanTrans(returnValue, freqItemP, indexList, nextChild->node, data, fpMap, isChoose, profit);
			
			nextChild = nextChild->next;
		}

		if (node->parent == NULL)
		{
			// root
			return;
		}
		else
		{
			doCreateFP(returnValue, freqItemP, indexList, node, fpMap, isChoose, profit);
		}
	}

}


/*
 * Function  : fp_scanTrans2
 * Purpose   : to scan the transactions in the database (for the calculation of H)
 * Parameters: 
 *   1. returnValue: the profit of the transaction
 *   2. freqItemP: the ordered freq. items
 *   3. indexList: the list of index of freq. item
 *   4. node: the root node of the FP-tree
 *   5. data: the data in the database
 *   6. fpMap: the mapping table from the header table to item no.
 *   7. isChoose: the array indicating whether the item is chosen or not
 *   8. profit: the profit of items
 *   9. revNo: the reverse mapping of fpMap
 *  10. itemSpecial: the item special for the calculation of H
 */
//fp_scanTrans2(&returnValue, &returnValue2, freqItemP, indexList, root, data, fpMap, isChoose, profit);
// for H
void fp_scanTrans2(double *returnValue, double *returnValue2, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, int *isChoose, int *profit, double *revNo, int itemSpecial)
{
	
	//int itemNo;

	childLink nextChild;
	
	if (node == NULL)
	{
		return;
	}


	if (node->children == NULL)
	{
		// this node has no children
		// (i.e. this node is leaf)
		
		
		//doCreateFP(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, short *fpMap, int *isChoose, int *profit)
		doCreateFP2(returnValue, returnValue2, freqItemP, indexList, node, fpMap, isChoose, profit, revNo, itemSpecial);
	}
	else
	{
		// this node has children
		// (need to explore the children first
		//  by using depth-first search)

		// explore children
		nextChild = node->children;
		while (nextChild != NULL)
		{
			//fp_scanTrans(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, int *isChoose, int *profit)
			fp_scanTrans2(returnValue, returnValue2, freqItemP, indexList, nextChild->node, data, fpMap, isChoose, profit, revNo, itemSpecial);
			
			nextChild = nextChild->next;
		}

		if (node->parent == NULL)
		{
			// root
			return;
		}
		else
		{
			doCreateFP2(returnValue, returnValue2, freqItemP, indexList, node, fpMap, isChoose, profit, revNo, itemSpecial);
		}
	}

}


/*
 * Function  : fp_scan_diff
 * Purpose   : to scan the transacions for the calculation of the difference of the profit 
 * Parameters: 
 *   1. returnValue: the profit of the transaction
 *   2. returnValue2: the difference of the profit of the transaction
 *   3. freqItemP: the ordered freq. items
 *   4. indexList: the list of index of freq. item
 *   5. node: the root node of the FP-tree
 *   6. data: the data in the database
 *   7. fpMap: the mapping table from the header table to item no.
 *   8. fpMap2: the mapping tabel from the header table to item no.
 *   9. isChoose: the array indicating whether the item is chosen or not
 *  10. profit: the profit of items
 *  11. itemAdd: the item just added 
 *  12. itemRemove: the item just removed
 */
// for diff

void fp_scanTrans_diff(double *returnValue, double *returnValue2, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, short *fpMap2, int *isChoose, int *profit, int itemAdd, int itemRemove)
{
	
	//int itemNo;

	childLink nextChild;
	
	if (node == NULL)
	{
		return;
	}


	if (node->children == NULL)
	{
		// this node has no children
		// (i.e. this node is leaf)
		
		
		//doCreateFP(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, short *fpMap, int *isChoose, int *profit)
		doCreateFP_diff(returnValue, returnValue2, freqItemP, indexList, node, fpMap, fpMap2, isChoose, profit, itemAdd, itemRemove);
	}
	else
	{
		// this node has children
		// (need to explore the children first
		//  by using depth-first search)

		// explore children
		nextChild = node->children;
		while (nextChild != NULL)
		{
			//fp_scanTrans(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, int *isChoose, int *profit)
			fp_scanTrans_diff(returnValue, returnValue2, freqItemP, indexList, nextChild->node, data, fpMap, fpMap2, isChoose, profit, itemAdd, itemRemove);
			
			nextChild = nextChild->next;
		}

		if (node->parent == NULL)
		{
			// root
			return;
		}
		else
		{
			doCreateFP_diff(returnValue, returnValue2, freqItemP, indexList, node, fpMap, fpMap2, isChoose, profit, itemAdd, itemRemove);
		}
	}

}


/******************************************************************************************
 * Function: buildTree()
 *
 * Description:
 *	Build the initial FP-tree.
 *
 * Invoked from:	
 *	main()
 *
 * Functions to be invoked:
 *	insert_tree()
 *	q_sortA()
 *
 * Global variables:
 *	root		-> Pointer to the root of this initial FP-tree
 *	headerTableLink	-> Header table for this initial FP-tree
 *
 * Global variables (read only):
 *	numLarge[]	-> Large k-itemsets resulting list for k = 1 to realK
 */
double buildTreeSet(transPtr *data, short *fpMap, int *isChoose, int *profit, double *revNo, int *chooseFlag)
{
 double returnValue = 0;
 int *freqItemP;	/* Store frequent items of a transaction */
 int *indexList;	/* indexList[i] = the index position in the large 1-item list storing freqItemP[i] */
 //int count;		/* Number of frequent items in a transaction */
 //FILE *fp;		/* Pointer to the database file */
 //int transSize;		/* Transaction size */
 //int item;		/* An item in the transaction */
 int i, j;//, m;
 //int path;		/* Number of new tree paths (i.e. new leaf nodes) created so far */

 //int itemNo;
 //int flag;


 returnValue = 0;

 /* Create header table */
 headerTableLinkSet = (FPTreeNode *) malloc (sizeof(FPTreeNode) * numLarge[0]);
 if (headerTableLinkSet == NULL) {
	printf("out of memory\n");
	exit(1);
 }
 for (i=0; i < numLarge[0]; i++)
	headerTableLinkSet[i] = NULL;
	
 /* Create root of the FP-tree */
 rootSet = (FPTreeNode) malloc (sizeof(FPNode));
 if (rootSet == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 /* Initialize the root node */
 rootSet->numPath = 1;
 rootSet->parent = NULL;
 rootSet->children = NULL;
 rootSet->hlink = NULL;

 /* Create freqItemP to store frequent items of a transaction */
 freqItemP = (int *) malloc (sizeof(int) * numItem);
 if (freqItemP == NULL) {
	printf("out of memory\n");
	exit(1);
 }	

 indexList = (int *) malloc (sizeof(int) * numItem);
 if (indexList == NULL) {
	printf("out of memory\n");
	exit(1);
 }	


#ifdef FP_SCAN
 //find_no(root);
 //scanf("%d", &flag);
	fp_copyCount(root);
	//void fp_scanTrans(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, int *isChoose, int *profit)
	fp_scanTrans(&returnValue, freqItemP, indexList, root, data, fpMap, isChoose, profit);
#else

	for (i = 0; i < noOfTrans; i++)
	{
		fp_doTransaction(&returnValue, data[i], freqItemP, indexList, rootSet, headerTableLinkSet, fpMap, isChoose2, profit, 1);
	}
#endif

	// calculate the benefit of items in FP-tree
	getBenefitFP(&returnValue, rootSet, isChoose, profit, revNo);
	

 /* scan DB and insert frequent items into the FP-tree */
/* if ((fp = fopen(dataFile, "r")) == NULL) {
        printf("Can't open data file, %s.\n", dataFile);
        exit(1);
 }


 for (i=0; i < numTrans; i++) {

	//* Read the transaction size 
	fscanf(fp, "%d", &transSize);

	count = 0;
 	path = 0;

	for (j=0; j < transSize; j++) {

		//* Read a transaction item 
		fscanf(fp, "%d", &item);

		//* Store the item to the frequent list, freqItemP, 
		//* if it is a large 1-item.
		// 
		for (m=0; m < numLarge[0]; m++) {
			if (item == largeItem1[m]) {
				// Store the item 
				freqItemP[count] = item;
				// Store the position in the large 1-itemset list storing this item 
				indexList[count] = m;
				count++;
				break;
			} 
		}
	}

	//* Sort the items in the frequent item list in ascending order of indexList,
	// * i.e. sort according to the order of the large 1-itemset list
	// 
	q_sortA(indexList, freqItemP, 0, count-1, count);

	// Insert the frequent patterns of this transaction to the FP-tree. 
	insert_tree(&(freqItemP[0]), &(indexList[0]), 1, 0, count, rootSet, headerTableLinkSet, &path);
 } 
 fclose(fp);
*/


 free(freqItemP);
 free(indexList);
 //free(largeItem1);
 //free(support1);
 free_fpSet();

 return returnValue;
}




/******************************************************************************************
 * Function: buildTree()
 *
 * Description:
 *	Build the initial FP-tree.
 *
 * Invoked from:	
 *	main()
 *
 * Functions to be invoked:
 *	insert_tree()
 *	q_sortA()
 *
 * Global variables:
 *	root		-> Pointer to the root of this initial FP-tree
 *	headerTableLink	-> Header table for this initial FP-tree
 *
 * Global variables (read only):
 *	numLarge[]	-> Large k-itemsets resulting list for k = 1 to realK
 */
double buildTreeSet2(transPtr *data, short *fpMap, int *isChoose, int *profit, double *revNo, int *chooseFlag, int itemSpecial)
{
 double returnValue = 0;
 double returnValue2 = 0;
 double returnValue3 = 0;
 int *freqItemP;	/* Store frequent items of a transaction */
 int *indexList;	/* indexList[i] = the index position in the large 1-item list storing freqItemP[i] */
 int count;		/* Number of frequent items in a transaction */
 //FILE *fp;		/* Pointer to the database file */
 //int transSize;		/* Transaction size */
 //int item;		/* An item in the transaction */
 int i, j;//, m;
 int path;		/* Number of new tree paths (i.e. new leaf nodes) created so far */

 int itemNo;
 int flag;

 double temp;
 int isExist;


 returnValue = 0;

 /* Create header table */
 headerTableLinkSet = (FPTreeNode *) malloc (sizeof(FPTreeNode) * numLarge[0]);
 if (headerTableLinkSet == NULL) {
	printf("out of memory\n");
	exit(1);
 }
 for (i=0; i < numLarge[0]; i++)
	headerTableLinkSet[i] = NULL;
	
 /* Create root of the FP-tree */
 rootSet = (FPTreeNode) malloc (sizeof(FPNode));
 if (rootSet == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 /* Initialize the root node */
 rootSet->numPath = 1;
 rootSet->parent = NULL;
 rootSet->children = NULL;
 rootSet->hlink = NULL;

 /* Create freqItemP to store frequent items of a transaction */
 freqItemP = (int *) malloc (sizeof(int) * numItem);
 if (freqItemP == NULL) {
	printf("out of memory\n");
	exit(1);
 }	

 indexList = (int *) malloc (sizeof(int) * numItem);
 if (indexList == NULL) {
	printf("out of memory\n");
	exit(1);
 }	

#ifdef FP_SCAN
 //find_no(root);
 //scanf("%d", &flag);
	fp_copyCount(root);
	//void fp_scanTrans(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, int *isChoose, int *profit)
	fp_scanTrans2(&returnValue, &returnValue2, freqItemP, indexList, root, data, fpMap, isChoose, profit, revNo, itemSpecial);
#else

	for (i = 0; i < noOfTrans; i++)
	{
		flag = isContainCase(data[i], isChoose);

		if (flag == 0)
		{
			// all unselected items

			// do nothing
		}
		else 
		{
			// check whether the trans contains specialItem
			isExist = 0;
			for (j = 0; j < data[i]->noOfItem; j++)
			{
				if (data[i]->item[j] == itemSpecial)
				{
					isExist = 1;
					break;
				}
			}

			if (flag == 2)
			{
				// all selected items

				temp = 0;
				
				for (j = 0; j < data[i]->noOfItem; j++)
				{
					itemNo = data[i]->item[j];
					
					temp += profit[itemNo];
				}
				returnValue += temp;

				if (isExist == 0)
				{
					returnValue2 += temp;
				}
				else
				{
					returnValue2 = returnValue2 + temp;  

					for (j = 0; j < data[i]->noOfItem; j++)
					{
						itemNo = data[i]->item[j];
						if (itemNo == itemSpecial)
						{
							returnValue2 -= profit[itemSpecial];
						}
						else
						{
							returnValue2 -= profit[itemNo]*getMatrix(itemSpecial, itemNo)*revNo[itemNo];
						}
					}
				}
			}
			else
			{
				// flag == 1
				// some selected and some unselected
				
				count = 0;
				path = 0;
				
				for (j = 0; j < data[i]->noOfItem; j++)
				{
					itemNo = data[i]->item[j];
					
					freqItemP[count] = itemNo;
					indexList[count] = (int)fpMap[itemNo];
					count++;
				}
				
				q_sortA(indexList, freqItemP, 0, count-1, count);
				
				insert_tree(freqItemP, indexList, 1, 0, count, rootSet, headerTableLinkSet, &path);
			}
		}
	}
#endif

	// calculate the benefit of items in FP-tree
	getBenefitFP2(&returnValue, &returnValue2, rootSet, isChoose, profit, revNo, itemSpecial);

	

 /* scan DB and insert frequent items into the FP-tree */
/* if ((fp = fopen(dataFile, "r")) == NULL) {
        printf("Can't open data file, %s.\n", dataFile);
        exit(1);
 }


 for (i=0; i < numTrans; i++) {

	//* Read the transaction size 
	fscanf(fp, "%d", &transSize);

	count = 0;
 	path = 0;

	for (j=0; j < transSize; j++) {

		//* Read a transaction item 
		fscanf(fp, "%d", &item);

		//* Store the item to the frequent list, freqItemP, 
		//* if it is a large 1-item.
		// 
		for (m=0; m < numLarge[0]; m++) {
			if (item == largeItem1[m]) {
				// Store the item 
				freqItemP[count] = item;
				// Store the position in the large 1-itemset list storing this item 
				indexList[count] = m;
				count++;
				break;
			} 
		}
	}

	//* Sort the items in the frequent item list in ascending order of indexList,
	// * i.e. sort according to the order of the large 1-itemset list
	// 
	q_sortA(indexList, freqItemP, 0, count-1, count);

	// Insert the frequent patterns of this transaction to the FP-tree. 
	insert_tree(&(freqItemP[0]), &(indexList[0]), 1, 0, count, rootSet, headerTableLinkSet, &path);
 } 
 fclose(fp);
*/


 free(freqItemP);
 free(indexList);
 //free(largeItem1);
 //free(support1);
 free_fpSet();

 returnValue3 = returnValue - returnValue2;

 return returnValue3;
}




/******************************************************************************************
 * Function: buildTree()
 *
 * Description:
 *	Build the initial FP-tree.
 *
 * Invoked from:	
 *	main()
 *
 * Functions to be invoked:
 *	insert_tree()
 *	q_sortA()
 *
 * Global variables:
 *	root		-> Pointer to the root of this initial FP-tree
 *	headerTableLink	-> Header table for this initial FP-tree
 *
 * Global variables (read only):
 *	numLarge[]	-> Large k-itemsets resulting list for k = 1 to realK
 */
double buildTreeSet_diff(transPtr *data, short *fpMap, short *fpMap2, int *isChoose, int *profit, double *revNo, int *chooseFlag, int itemAdd, int itemRemove)
{
 double returnValue = 0;
 double returnValue2 = 0;
 double returnValue3 = 0;
 int *freqItemP;	/* Store frequent items of a transaction */
 int *indexList;	/* indexList[i] = the index position in the large 1-item list storing freqItemP[i] */
 //int count;		/* Number of frequent items in a transaction */
 //FILE *fp;		/* Pointer to the database file */
 //int transSize;		/* Transaction size */
 //int item;		/* An item in the transaction */
 int i, j;//, m;
 //int path;		/* Number of new tree paths (i.e. new leaf nodes) created so far */

 //int itemNo;
 //int flag;

 int *isChoose2;


 returnValue = 0;

// for set1
 /* Create header table */
 headerTableLinkSet = (FPTreeNode *) malloc (sizeof(FPTreeNode) * numLarge[0]);
 if (headerTableLinkSet == NULL) {
	printf("out of memory\n");
	exit(1);
 }
 for (i=0; i < numLarge[0]; i++)
	headerTableLinkSet[i] = NULL;
	
 /* Create root of the FP-tree */
 rootSet = (FPTreeNode) malloc (sizeof(FPNode));
 if (rootSet == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 /* Initialize the root node */
 rootSet->numPath = 1;
 rootSet->parent = NULL;
 rootSet->children = NULL;
 rootSet->hlink = NULL;



// for set2
 /* Create header table */
 headerTableLinkSet2 = (FPTreeNode *) malloc (sizeof(FPTreeNode) * numLarge[0]);
 if (headerTableLinkSet2 == NULL) {
	printf("out of memory\n");
	exit(1);
 }
 for (i=0; i < numLarge[0]; i++)
	headerTableLinkSet2[i] = NULL;
	
 /* Create root of the FP-tree */
 rootSet2 = (FPTreeNode) malloc (sizeof(FPNode));
 if (rootSet2 == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 /* Initialize the root node */
 rootSet2->numPath = 1;
 rootSet2->parent = NULL;
 rootSet2->children = NULL;
 rootSet2->hlink = NULL;




 /* Create freqItemP to store frequent items of a transaction */
 freqItemP = (int *) malloc (sizeof(int) * numItem);
 if (freqItemP == NULL) {
	printf("out of memory\n");
	exit(1);
 }	

 indexList = (int *) malloc (sizeof(int) * numItem);
 if (indexList == NULL) {
	printf("out of memory\n");
	exit(1);
 }	


#ifdef FP_SCAN
 //find_no(root);
 //scanf("%d", &flag);
	fp_copyCount(root);
	//void fp_scanTrans(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, transPtr *data, short *fpMap, int *isChoose, int *profit)
	fp_scanTrans_diff(&returnValue, &returnValue2, freqItemP, indexList, root, data, fpMap, fpMap2, isChoose, profit, itemAdd, itemRemove);
#else

	for (i = 0; i < noOfTrans; i++)
	{
		printf("Problem!!!!.....\n"); //problem:need to calculate returnValue2
		fp_doTransaction(&returnValue, data[i], freqItemP, indexList, rootSet, headerTableLinkSet, fpMap, isChoose2, profit, 1);
	}
#endif

	// calculate the benefit of items in FP-tree

	isChoose2 = (int *) malloc(sizeof(int)*noOfItem);
	memcpy(isChoose2, isChoose, sizeof(int)*noOfItem);

	// for S U itemAdd
	isChoose2[itemAdd] = 1;
	isChoose2[itemRemove] = 0;
	getBenefitFP(&returnValue, rootSet, isChoose2, profit, revNo);

	// for S U itemRemove
	isChoose2[itemAdd] = 0;
	isChoose2[itemRemove] = 1;
	getBenefitFP(&returnValue2, rootSet2, isChoose2, profit, revNo);

	

 /* scan DB and insert frequent items into the FP-tree */
/* if ((fp = fopen(dataFile, "r")) == NULL) {
        printf("Can't open data file, %s.\n", dataFile);
        exit(1);
 }


 for (i=0; i < numTrans; i++) {

	//* Read the transaction size 
	fscanf(fp, "%d", &transSize);

	count = 0;
 	path = 0;

	for (j=0; j < transSize; j++) {

		//* Read a transaction item 
		fscanf(fp, "%d", &item);

		//* Store the item to the frequent list, freqItemP, 
		//* if it is a large 1-item.
		// 
		for (m=0; m < numLarge[0]; m++) {
			if (item == largeItem1[m]) {
				// Store the item 
				freqItemP[count] = item;
				// Store the position in the large 1-itemset list storing this item 
				indexList[count] = m;
				count++;
				break;
			} 
		}
	}

	//* Sort the items in the frequent item list in ascending order of indexList,
	// * i.e. sort according to the order of the large 1-itemset list
	// 
	q_sortA(indexList, freqItemP, 0, count-1, count);

	// Insert the frequent patterns of this transaction to the FP-tree. 
	insert_tree(&(freqItemP[0]), &(indexList[0]), 1, 0, count, rootSet, headerTableLinkSet, &path);
 } 
 fclose(fp);
*/



 free(isChoose2);
 free(freqItemP);
 free(indexList);
 //free(largeItem1);
 //free(support1);
 free_fpSet();
 free_fpSet2();

 returnValue3 = returnValue - returnValue2;
 return returnValue3;
}




/*
 * Function  : free_fpSet
 * Purpose   : to free the fp-tree
 * Parameters: 
 *   1. none
 */
void free_fpSet()
{
	free(headerTableLinkSet);
	destroyTree(rootSet);
}


/*
 * Function  : free_fpSet2
 * Purpose   : to free the fp-tree
 * Parameters: 
 *   1. none
 */
void free_fpSet2()
{
	free(headerTableLinkSet2);
	destroyTree(rootSet2);
}


/*
 * Function  : getBenefitFP
 * Purpose   : to calculate the total profit of the selection
 * Parameters: 
 *   1. returnValue: the profit
 *   2. node: the root node of the fp-tree
 *   3. isChoose: the array indicating whether it is selected or not
 *   4. profit: the profit of items
 *   5. revNo: the reverse mapping of fpMap
 */
void getBenefitFP(double *returnValue, FPTreeNode node, int *isChoose, int *profit, double *revNo)
{
	int itemNo;

	int countNo;

	childLink nextChild;

	if (node->children == NULL)
	{
		// this node has no children
		// (i.e. this node is leaf)


		// check whether this item is selected
		itemNo = node->item;
		if ((node->parent != NULL) && (isChoose[itemNo] == 1))
		{
			countNo = node->count;
			if (countNo == 1)
			{
				(*returnValue) += profit[itemNo]*(1-countNo*revNo[itemNo]);
			}
			else
			{
				(*returnValue) += profit[itemNo]*(1-countNo*revNo[itemNo])*countNo;
			}
		}
	}
	else
	{
		// this node has children
		// (need to explore the children first
		//  by using depth-first seearch)

		// explore children
		nextChild = node->children;
		while (nextChild != NULL)
		{
			getBenefitFP(returnValue, nextChild->node, isChoose, profit, revNo);

			nextChild = nextChild->next;
		}

		//void getBenefitFP(double *returnValue, FPTreeNode node, int *isChoose, int *profit, double *revNo)

		// explore parent
		// check whether this item is selected
		itemNo = node->item;
		if ((node->parent != NULL) && (isChoose[itemNo] == 1))
		{
			countNo = node->count;
			if (countNo == 1)
			{
				(*returnValue) += profit[itemNo]*(1-countNo*revNo[itemNo]);
			}
			else
			{
				(*returnValue) += profit[itemNo]*(1-countNo*revNo[itemNo])*countNo;
			}
		}
	}
}


/*
 * Function  : getBenefitFP2
 * Purpose   : to calculate the total profit of the selection
 * Parameters: 
 *   1. returnValue: the profit
 *   2. node: the root node of the fp-tree
 *   3. isChoose: the array indicating whether it is selected or not
 *   4. profit: the profit of items
 *   5. revNo: the reverse mapping of fpMap
 *   6. itemSpecial: the item to be treated
 */
void getBenefitFP2(double *returnValue, double *returnValue2, FPTreeNode node, int *isChoose, int *profit, double *revNo, int itemSpecial)
{
	int itemNo;

	int countNo;

	childLink nextChild;

	double temp;

	if (node->children == NULL)
	{
		// this node has no children
		// (i.e. this node is leaf)


		// check whether this item is selected
		itemNo = node->item;
		if ((node->parent != NULL) && (isChoose[itemNo] == 1))
		{
			countNo = node->count;
			temp = 0;
			if (countNo == 1)
			{
				temp += profit[itemNo]*(1-countNo*revNo[itemNo]);
			}
			else
			{
				temp += profit[itemNo]*(1-countNo*revNo[itemNo])*countNo;
			}
			(*returnValue) += temp;

			if (itemNo != itemSpecial)
			{
				(*returnValue2) += temp;
			}
		}
	}
	else
	{
		// this node has children
		// (need to explore the children first
		//  by using depth-first seearch)

		// explore children
		nextChild = node->children;
		while (nextChild != NULL)
		{
			//void getBenefitFP2(double *returnValue, double *returnValue2, FPTreeNode node, int *isChoose, int *profit, double *revNo, int itemSpecial)
			getBenefitFP2(returnValue, returnValue2, nextChild->node, isChoose, profit, revNo, itemSpecial);

			nextChild = nextChild->next;
		}

		//void getBenefitFP(double *returnValue, FPTreeNode node, int *isChoose, int *profit, double *revNo)

		// explore parent
		// check whether this item is selected
		itemNo = node->item;
		if ((node->parent != NULL) && (isChoose[itemNo] == 1))
		{
			countNo = node->count;
			temp = 0;
			if (countNo == 1)
			{
				temp += profit[itemNo]*(1-countNo*revNo[itemNo]);
			}
			else
			{
				temp += profit[itemNo]*(1-countNo*revNo[itemNo])*countNo;
			}

			(*returnValue) += temp;

			if (itemNo != itemSpecial)
			{
				(*returnValue2) += temp;
			}
		}
	}
}




